<?php

namespace Drupal\declarative_form_ajax;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InsertCommand;
use Drupal\Core\Ajax\PrependCommand;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\RenderElement;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the form AJAX handlers for declarative AJAX.
 */
class FormAjax {

  /**
   * After build callback to set up declarative AJAX.
   *
   * This should be set as an '#after_build' callback on the whole form, for
   * example:
   *
   * @code
   * $form['#after_build'][] = FormAjax::class .  '::ajaxAfterBuild';
   * @endcode
   *
   * This walks the entire form, looking for elements with
   * ['#ajax']['updated_by'] set. It then sets up the targeted elements to be
   * AJAX triggers.
   *
   * This needs to do a few hacks because it's operating outside of core
   * FormAPI:
   *  - AJAX has already been processed by this point, so we need to forcibly
   *    process it again.
   *  - In an AJAX request, the form build process calls setTriggeringElement()
   *    before #after_build callbacks are executed, and does not set it by
   *    reference, and so we need to update the form state's copy so it has our
   *    callback.
   *  - Form elements which have 'updated_by' and no other #ajax properties will
   *    nonetheless be processed by RenderElement as if they need to trigger an
   *    AJAX request. This needs to be undone here.
   */
  public static function ajaxAfterBuild($form, FormStateInterface $form_state) {
    $ajax_controlling_elements = [];

    // Find all elements which are the target of an AJAX updating element.
    Element::walkChildrenRecursive($form, function(&$walk_element) use (&$form, $form_state, &$ajax_controlling_elements) {
      if (!isset($walk_element['#ajax']['updated_by'])) {
        return;
      }

      foreach ($walk_element['#ajax']['updated_by'] as $address) {
        // Use a key based on the element address so we only process each
        // element once. This is important so we retain any existing AJAX
        // callback on an element.
        $ajax_controlling_elements[implode(':', $address)] = $address;
      }

      // Drupal\Core\Render\Element\RenderElement::preRenderAjaxForm() will set
      // the element up for AJAX triggering even if the element only seeks to
      // declare itself as a dependency, with no triggering behaviour of its
      // own. We need to undo this, otherwise clicking it will trigger a
      // pointless AJAX request.
      // This is only necessary because we're operating outside of core's AJAX
      // handling.
      if (count($walk_element['#ajax']) == 2 && isset($walk_element['#ajax']['event'])) {
        unset($walk_element['#ajax']['event']);
        unset($walk_element['#attached']['drupalSettings']['ajax']);
        unset($walk_element['#attached']['drupalSettings']['ajaxTrustedUrl']);
      }
    });

    // Set up each triggering element.
    foreach ($ajax_controlling_elements as $address) {
      $triggering_element =& NestedArray::getValue($form, $address);

      if (!isset($triggering_element['#ajax_processed'])) {
        // @todo: Throw an exception, this triggering element won't work with
        // AJAX.
      }
      elseif (($triggering_element['#ajax_processed']) == FALSE) {
        $triggering_element['#ajax']['callback'] = static::class . '::ajaxCallback';

        // The element was already processed for AJAX by
        // RenderElement::processAjaxForm() but there were no AJAX settings at
        // the time, so we need to send it through again.
        unset($triggering_element['#ajax_processed']);
        $triggering_element = RenderElement::processAjaxForm($triggering_element, $form_state, $form);

        // During an AJAX request, the triggering element is set on the form
        // state earlier than the #after_build process, and so will not have
        // our callback. Therefore, we need to set it again.
        if ($current_triggering_element =& $form_state->getTriggeringElement()) {
          if ($current_triggering_element['#array_parents'] == $triggering_element['#array_parents']) {
            $current_triggering_element['#ajax']['callback'] = static::class . '::ajaxCallback';
          }
        }
      }
      elseif (($triggering_element['#ajax_processed']) == TRUE) {
        // The element has its own AJAX callback, which we need to register so
        // that we call it from our own callback.
        $triggering_element['#ajax']['prior_callback'] = $triggering_element['#ajax']['callback'];

        $triggering_element['#ajax']['callback'] = static::class . '::ajaxCallback';

        // During an AJAX request, the triggering element is set on the form
        // state earlier than the #after_build process, and so will not have
        // our callback. Therefore, we need to set it again.
        if ($current_triggering_element =& $form_state->getTriggeringElement()) {
          if ($current_triggering_element['#array_parents'] == $triggering_element['#array_parents']) {
            $current_triggering_element['#ajax']['prior_callback'] = $current_triggering_element['#ajax']['callback'];

            $current_triggering_element['#ajax']['callback'] = static::class . '::ajaxCallback';
          }
        }
      }
    }

    return $form;
  }

  /**
   * Form AJAX callback.
   *
   * This is set as an AJAX callback on AJAX elements by
   * static::ajaxAfterBuild().
   *
   * If an AJAX element already had an AJAX callback, this takes care of calling
   * it and adds our response to its result.
   *
   * @param array &$form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request object.
   */
  public static function ajaxCallback(&$form, FormStateInterface $form_state, Request $request): AjaxResponse {
    $triggering_element = $form_state->getTriggeringElement();
    $triggering_element_parents = $triggering_element['#array_parents'];

    // Call a prior callback that we've replaced.
    // We take its response and use that, as it's not possible to copy AJAX
    // commands from one AjaxResponse to another.
    if (isset($triggering_element['#ajax']['prior_callback'])) {
      $prior_callback = $triggering_element['#ajax']['prior_callback'];
      $prior_callback = $form_state->prepareCallback($prior_callback);

      // Make a copy of the form array, as the callback apparently messes with
      // it.
      $form_copy = $form;
      $prior_callback_result = call_user_func_array($prior_callback, [&$form_copy, &$form_state, $request]);

      if ($prior_callback_result instanceof AjaxResponse) {
        $response = $prior_callback_result;
      }
      else {
        $route_match = \Drupal::routeMatch();

        $response = \Drupal::service('main_content_renderer.ajax')->renderResponse($prior_callback_result, $request, $route_match);
      }
    }
    else {
      $response = new AjaxResponse();
    }

    // An array of elements from the form that should be rendered and returned
    // in the response.
    $collected_elements = [];

    // Walk the entire form recursively, looking for elements which say they
    // update on the triggering element.
    Element::walkChildrenRecursive($form, function($element) use ($triggering_element_parents, &$collected_elements) {
      if (!isset($element['#ajax']['updated_by'])) {
        return;
      }

      foreach ($element['#ajax']['updated_by'] as $updated_by) {
        if ($updated_by == $triggering_element_parents) {
          $collected_elements[] = $element;
        }
      }
    });

    // Render each element that should update and add it to the response.
    foreach ($collected_elements as $updated_element) {
      // If the element is part of the group (#group is set on it) it won't be
      // rendered unless we remove #group from it. This is caused by
      // \Drupal\Core\Render\Element\RenderElement::preRenderGroup(), which
      // prevents all members of groups from being rendered directly.
      if (!empty($updated_element['#group'])) {
        unset($updated_element['#group']);
      }

      $html = \Drupal::service('renderer')->renderRoot($updated_element);
      $response->addAttachments($updated_element['#attached']);

      $element_data_selector = $updated_element['#attributes']['data-drupal-selector'];
      if (in_array($updated_element['#type'], ['details', 'container'])) {
        // Container elements ('details' and 'container') have their
        // data-drupal-selector attribute on a wrapping DIV.
        $insert_selector = "*[data-drupal-selector=\"{$element_data_selector}\"]";
      }
      else {
        // Other elements only have their data-drupal-selector attribute on the
        // input element itself, and so we have to rely on the CSS 'form-item'
        // class to get the wrapping DIV.
        // WARNING: This may not work on all themes.
        $insert_selector = "div.form-item:has(*[data-drupal-selector=\"{$element_data_selector}\"])";
      }

      $response->addCommand(new InsertCommand($insert_selector, $html));
    }

    // Output status messages.
    $status_messages = ['#type' => 'status_messages'];
    $output = \Drupal::service('renderer')->renderRoot($status_messages);
    if (!empty($output)) {
      // Prepend the status messages to the triggering element.
      $triggering_element_data_selector = $triggering_element['#attributes']['data-drupal-selector'];
      $messages_selector = "div.form-item:has(*[data-drupal-selector=\"{$triggering_element_data_selector}\"])";

      // @todo We can't use NULL as a selector.
      // See https://www.drupal.org/project/drupal/issues/3440854
      $response->addCommand(new PrependCommand($messages_selector, $output));
    }

    return $response;
  }

}
